package pathfinding;

import game.PixelData;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

public class Area {
	public ArrayList<ArrayList<PathNode>> nodeMap;
	public Navigator navigator;
	public int width, height;
	
	private List<PathNode> open = new ArrayList<PathNode>();
	private List<PathNode> closed = new ArrayList<PathNode>();
	private PathNode start;
	private PathNode goal;
	private Heuristic heuristic;
	
	public Area(int w, int h) {
		initialize(w, h);
	}
	
	public void initialize(int w, int h) {
		this.width = w;
		this.height = h;
		this.start = null;
		this.goal = null;
		this.nodeMap = new ArrayList<ArrayList<PathNode>>();
		this.heuristic = new DiagonalHeuristic();
		
		for (int y = 0; y < h; y++) {
			ArrayList<PathNode> temp = new ArrayList<PathNode>();
			for (int x = 0; x < w; x++) {
				PathNode node = new PathNode(x, y);
				temp.add(node);
			}
			nodeMap.add(temp);
		}
	}
	
	public ArrayList<PathNode> createPath() {
		start.distanceFromStart = 0;
		PathNode current = start;
		open.clear();
		closed.clear();
		
		open.add(start);
		
		while (!open.isEmpty()) {
			Collections.sort(open);
			current = open.remove(0);
			if (current.equals(goal)) {
				goal.parent = current;
				goal.distanceFromStart = current.distanceFromStart + getDistanceBetween(current, goal);
				goal.heuristicDistanceToGoal = 0;
				break;
			}
			open.remove(current);
			closed.add(current);
			ArrayList<PathNode> neighbors = getNeighbors(current);
			if (neighbors == null)
				break;
			for (PathNode neighbor : neighbors) {
				boolean neighborIsBetter;
				if (closed.contains(neighbor))
					continue;
				double neighborDistanceFromStart = current.distanceFromStart + getDistanceBetween(current, neighbor);
				double neighborDistanceToGoal = getEstimatedDistanceToGoal(neighbor, goal);
				double currentDistanceToGoal = getEstimatedDistanceToGoal(current, goal);
				if (neighborDistanceToGoal >= currentDistanceToGoal)
					continue;
				if (!neighbor.isObstacle()) {
					if (!open.contains(neighbor)) {
						open.add(neighbor);
						Collections.sort(open);
						neighborIsBetter = true;
					}
					else if (neighborDistanceFromStart < current.distanceFromStart)
						neighborIsBetter = true;
					else
						neighborIsBetter = false;
					if (neighborIsBetter) {
						neighbor.parent = current;
						neighbor.distanceFromStart = neighborDistanceFromStart;
						neighbor.heuristicDistanceToGoal = getEstimatedDistanceToGoal(neighbor, goal);
					}
				}
				
			}
		}
		return reconstructPath(current);
	}
	
	private ArrayList<PathNode> getCardinalNeighbors(PathNode node) {
		ArrayList<PathNode> neighborList = new ArrayList<PathNode>();
		int x = node.x;
		int y = node.y;
		
		int x0 = x - 1;
		int x2 = x + 1;
		int y0 = y - 1;
		int y2 = y + 1;
		
		if (x < 0 || y < 0 || x > width - 1 || y > height - 1)
			return null;
		
		//----------------------------------------------
		if (x0 > 0)
			neighborList.add(nodeMap.get(y).get(x0));
		//----------------------------------------------
		if (y0 > 0)
			neighborList.add(nodeMap.get(y0).get(x));
		if (y2 < height - 1)
			neighborList.add(nodeMap.get(y2).get(x));
		//----------------------------------------------
		if (x2 < width - 1)
			neighborList.add(nodeMap.get(y).get(x2));
		//----------------------------------------------
		
		return neighborList;
	}
	
	private ArrayList<PathNode> getNeighbors(PathNode node) {
		ArrayList<PathNode> neighborList = new ArrayList<PathNode>();
		int x = node.x;
		int y = node.y;
		
		int x0 = x - 1;
		int x2 = x + 1;
		int y0 = y - 1;
		int y2 = y + 1;
		
		if (x < 0 || y < 0 || x > width - 1 || y > height - 1)
			return null;
		
		//----------------------------------------------
		if (x0 > 0 && y0 > 0)
			neighborList.add(nodeMap.get(y0).get(x0));
		if (x0 > 0)
			neighborList.add(nodeMap.get(y).get(x0));
		if (x0 > 0 && y2 < height - 1)
			neighborList.add(nodeMap.get(y2).get(x0));
		//----------------------------------------------
		if (y0 > 0)
			neighborList.add(nodeMap.get(y0).get(x));
		if (y2 < height - 1)
			neighborList.add(nodeMap.get(y2).get(x));
		//----------------------------------------------
		if (x2 < width - 1 && y0 > 0)
			neighborList.add(nodeMap.get(y0).get(x2));
		if (x2 < width - 1)
			neighborList.add(nodeMap.get(y).get(x2));
		if (x2 < width - 1 && y2 < height - 1)
			neighborList.add(nodeMap.get(y2).get(x2));
		//----------------------------------------------
		
		return neighborList;
	}
	
	private double getDistanceBetween(PathNode n1, PathNode n2) {
		//Current node and neighbor node, each of them are placed next to each other. Thus, the cost
		//from current node to neighbor node is 1.
		if (n1.x == n2.x || n1.y == n2.y)
			return 1.0;
		//Current node and neighbor node, each of them are placed diagonally to each other. The cost
		//from current node to neighbor node is the diagonal cost.
		return Math.hypot(1, 1);
	}
	
	public ArrayList<PathNode> reconstructPath(PathNode node) {
		ArrayList<PathNode> path = new ArrayList<PathNode>();
		while (node.parent != null) {
			path.add(node);
			node = node.parent;
		}
		if (!path.isEmpty())
			return path;
		return null;
	}
	
	public void setStartNode(PixelData data) {
		this.start = new PathNode(data.x, data.y);
	}
	
	public void setGoalNode(PixelData data) {
		this.goal = new PathNode(data.x, data.y);
	}
	
	public void reset() {
		this.start = null;
		this.goal = null;
		
		for (int y = 0; y < this.height; y++) {
			for (int x = 0; x < this.width; x++) {
				PathNode node = this.nodeMap.get(y).get(x);
				node.distanceFromStart = 0;
				node.heuristicDistanceToGoal = 0;
				node.parent = null;
			}
		}
	}
	
	private double getEstimatedDistanceToGoal(PathNode start, PathNode finish) {
		double h_diagonal = Math.min(Math.abs(start.x - finish.x), Math.abs(start.y - finish.y));
		double h_straight = Math.abs(start.x - finish.x) + Math.abs(start.y - finish.y);
		double h_result = Math.sqrt(2) * h_diagonal + (h_straight - 2 * h_diagonal);
		
		//breaking ties
		double p = 1 / 10000;
		h_result *= (1.0 + p);
		return h_result;
	}
}
